==> japhar-0.04.tar.gz: japhar-0.04/lib/libruntime/msgc.c <==
/* -*- Mode: C; c-file-style: "gnu" -*-
   msgc.c -- A simple, non-concurrent mark and sweep collector.
   Created: Chris Toshok <toshok@hungry.com>, 10-Aug-1997
*/
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "jniint.h"
#include "qop.h"
#include "gc.h"
#include "objects.h"
#include "arrays.h"
#include "array-class.h"
#include "ClazzFile.h"
#include "native-threads.h"

#include <stdlib.h>

#define MAX_BYTES 32000

GC_state*
GC_createGCState(JavaVM *vm)
{
  GC_state *state = (GC_state*)calloc(1, sizeof(GC_state));

  state->mon = MONITOR_create();
  state->vm = vm;

  return state;
}

void
GC_deleteGCState(GC_state* state)
{
  if (state->finalizer_q)
    GC_runFinalizers(state->vm);

  /* run the finalizers on all the roots. */

  /* destroy the monitor associated with this GC_state. */
  MONITOR_destroy(state->mon);

  /* free up the GC_state memory. */
  free(state);
}

void
GC_finalize(JavaVM *vm, GC_obj *gc_obj)
{
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;

  if (gc_obj->gc_flags & GC_FLAG_CLASS
      && !state->class_gc_enabled)
    return;

  if (gc_obj->gc_flags & GC_FLAG_OBJECT)
    state->total_num_objects_finalized++;
  else if (gc_obj->gc_flags & GC_FLAG_ARRAY)
    state->total_num_arrays_finalized++;
  else
    state->total_num_classes_finalized++;

  state->total_num_bytes_collected += gc_obj->size;

  (*gc_obj->finalize)(gc_obj);
}

void
GC_deleteGlobal(JavaVM *vm, GC_obj *gc_obj)
{
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;
  MONITOR_enter(state->mon);

  UNQUEUE(gc_obj, state->global_refs);
  GC_finalize(vm, gc_obj);

  MONITOR_exit(state->mon);
}

void
GC_deleteThreadLocal(JavaVM* vm, GC_obj *gc_obj)
{
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;
  MONITOR_enter(state->mon);
  ENQUEUE(gc_obj, state->finalizer_q);
  MONITOR_exit(state->mon);
}

void
GC_runFinalizers(JavaVM *vm)
{
  HungryJavaVM *hvm = (HungryJavaVM*)vm;
  GC_state *state = hvm->_gc_state;
  int classes, objs, arrays;

  classes =
    objs =
    arrays = 0;

  state->verbose_gc = hvm->_args.enableVerboseGC;
  state->class_gc_enabled = hvm->_args.enableClassGC;

  state->num_finalizer_runs++;

  if (state->verbose_gc)
    printf ("GC: finalizing.\n");

  while(state->finalizer_q)
    {
      GC_obj *next = state->finalizer_q->next;
      GC_obj *cur = state->finalizer_q;

      UNQUEUE(cur, state->finalizer_q);
      if (cur->gc_flags & GC_FLAG_OBJECT)
	objs ++;
      else if (cur->gc_flags & GC_FLAG_ARRAY)
	arrays++;
      else
	classes++;

      GC_finalize(vm, cur);

      state->finalizer_q = next;
    }

  if (state->verbose_gc)
    printf ("GC: finalizing complete (%d objs/%d classes/%d arrays)\n", objs, classes, arrays);
}

static void
gc_markStuff(JavaVM *vm, GC_obj *gc_obj)
{
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;

  MONITOR_enter(state->mon);

  if (gc_obj->gc_flags & GC_FLAG_OBJECT)
    {
      obj_thunk *obj = (obj_thunk*)gc_obj;
      ClazzFile *clazz;
	  
      while (!IS_POST(obj))
	obj --;
      obj ++;
	  
      if (((GC_obj*)obj)->gc_flags & GC_FLAG_MARKED)
	return;
      ((GC_obj*)obj)->gc_flags |= GC_FLAG_MARKED;

      state->num_objects_reachable ++;

      while (!IS_POST(obj))
	{
	  int i;

	  /* mark the class of this thunk */
	  clazz = (ClazzFile*)STATIC_DATA(obj);
	  gc_markStuff(vm, (GC_obj*)clazz);

	  /* now go through the fields, and mark all the objects. */
	  for (i = 0; i < clazz->num_instance_fields; i ++)
	    {
	      if (INSTANCE_DATA(obj)->fields[i].tag == SIG_JOBJECT
		  && INSTANCE_DATA(obj)->fields[i].value.l != NULL)
		gc_markStuff(vm, (GC_obj*)INSTANCE_DATA(obj)->fields[i].value.l);
	    }

	  obj++;
	}
    }
  else if (gc_obj->gc_flags & GC_FLAG_ARRAY)
    {
      JavaStackItem type_item;

      if (gc_obj->gc_flags & GC_FLAG_MARKED)
	return;
      gc_obj->gc_flags |= GC_FLAG_MARKED;
	  
      state->num_arrays_reachable ++;

      get_instance_field((jobject)gc_obj,
			 STATIC_DATA((jobject)gc_obj)->fields[ARRAY_TYPE_INDEX],
			 &type_item);

      if (type_item.value.i == SIG_JOBJECT)
	{
	  JavaStackItem length_item;
	  JavaStackItem body_item;
	  int i;

	  get_instance_field((jobject)gc_obj,
			     STATIC_DATA((jobject)gc_obj)->fields[ARRAY_LENGTH_INDEX],
			     &length_item);
		  
	  get_instance_field((jobject)gc_obj,
			     STATIC_DATA((jobject)gc_obj)->fields[ARRAY_BODY_INDEX],
			     &body_item);

	  for (i = 0; i < length_item.value.i; i ++)
	    {
	      if (((jobject*)body_item.value.l)[i])
		gc_markStuff(vm, (GC_obj*)((jobject*)body_item.value.l)[i]);
	    }
	}
    }
  else if (gc_obj->gc_flags & GC_FLAG_CLASS)
    {
      ClazzFile *cf = (ClazzFile*)gc_obj;
      int i;
	 
      if (gc_obj->gc_flags & GC_FLAG_MARKED)
	{
	  MONITOR_exit(state->mon);

	  return;
	}

      gc_obj->gc_flags |= GC_FLAG_MARKED;
 
      state->num_classes_reachable ++;

      /* here we need to mark all the objects that we have in static
	 fields */
      for (i = 0; i < cf->num_fields; i ++)
	{
	  FieldStruct *field = cf->fields[i];
		  
	  if (field->access_flags & ACC_STATIC
	      && ((field->sig->any.tag == SIG_PRIM
		   && field->sig->prim.type == SIG_JOBJECT)
		  || field->sig->any.tag == SIG_ARRAY))
	    {
	      /* so, the field we're looking at is either an object or
		 an array. */
	      obj_thunk *next_obj = cf->static_fields[ field->field_index ].value.l;
			  
	      gc_markStuff(vm, (GC_obj*)next_obj);
	    }
	}
    }

  MONITOR_exit(state->mon);
}

static void
gc_markRoots(JavaVM *vm)
{
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;
  GC_obj *cur;

  for (cur = state->roots;
       cur != NULL;
       cur = cur->next)
    {
      gc_markStuff(vm, cur);
    }
}

static void
gc_markFromRegisters(JavaVM *vm)
{
}

static void
gc_markOperandStack(JavaVM *vm,
		    OpStack *opstack)
{
  JavaStackItem *item;

  for (item = opstack->stack_top;
       item >= opstack->stack_bottom;
       item --)
    {
      if (item->tag == SIG_JOBJECT
	  && item->value.l != NULL)
	gc_markStuff(vm, (GC_obj*)item->value.l);
    }
}

static void
gc_markThreadReferences(JavaVM *vm)
{
  HungryJavaVM* hvm = (HungryJavaVM*)vm;
  HungryJNIEnv* env;

  for (env = hvm->_envs;
       env != NULL;
       env = env->next)
    {
      GC_obj *ref;
	  
      for (ref = env->_local_refs;
	   ref != NULL;
	   ref = ref->next)
	gc_markStuff(vm, ref);

      gc_markOperandStack(vm, env->_java_info->op_stack);
    }
}

void
GC_runCollector(JavaVM *vm)
{
  HungryJavaVM *hvm = (HungryJavaVM*)vm;
  GC_state *state = hvm->_gc_state;
  GC_obj *cur;
  int num_added_to_finalizer_q = 0;

  MONITOR_enter(state->mon);

  state->verbose_gc = hvm->_args.enableVerboseGC;
  state->class_gc_enabled = hvm->_args.enableClassGC;

  state->num_gcs++;

  state->num_objects_reachable =
    state->num_arrays_reachable =
    state->num_classes_reachable = 0;

  if (state->verbose_gc)
    printf ("<GC: starting (Mark and Sweep)>\n");

  /* first we go along our global ref's and mark the all unreached. */
  for (cur = state->global_refs;
       cur != NULL;
       cur = cur->next)
    {
      cur->gc_flags &= ~GC_FLAG_MARKED;
    }

  if (state->verbose_gc)
    printf ("  Mark phase.\n");
  /* mark phase */
  if (state->verbose_gc)
    printf ("    from roots.\n");
  gc_markRoots(vm);

  if (state->verbose_gc)
    printf ("    from registers.\n");
  gc_markFromRegisters(vm);

  if (state->verbose_gc)
    printf ("    from thread stacks.\n");
  gc_markThreadReferences(vm);

  if (state->verbose_gc)
    printf ("  Mark phase complete.\n");

  if (state->verbose_gc)
    {
      printf ("  %d objects reachable.\n",	state->num_objects_reachable);
      printf ("  %d arrays reachable.\n",	state->num_arrays_reachable);
      printf ("  %d classes reachable.\n",	state->num_classes_reachable);
    }

  if (state->verbose_gc)
    printf ("  Sweep phase.\n");
  /* sweep phase */
  cur = state->global_refs;
  while (cur != NULL)
    {
      if (!(cur->gc_flags & GC_FLAG_MARKED)
	  && (!(cur->gc_flags & GC_FLAG_CLASS) || state->class_gc_enabled))
	{
	  GC_obj *next = cur->next;
	  /* it's garbage. */
	  UNQUEUE(cur, state->global_refs);
	  ENQUEUE(cur, state->finalizer_q);

	  num_added_to_finalizer_q ++;
	  cur = next;
	}
      else
	cur = cur->next;
    }
  if (state->verbose_gc)
    printf ("  Sweep phase complete (%d added to finalizer queue.)\n",
	    num_added_to_finalizer_q);

  if (state->verbose_gc)
    GC_printStats(vm);

  GC_runFinalizers(vm);

  state->interval_num_bytes_allocated =
    state->interval_num_classes_allocated =
    state->interval_num_arrays_allocated =
    state->interval_num_objects_allocated = 0;

  MONITOR_exit(state->mon);
}

void
GC_printStats(JavaVM *vm)
{
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;

  MONITOR_enter(state->mon);

  printf ("\n");
  printf ("Number of GC's:\t%d\n",	state->num_gcs);
  printf ("GC Heap bytes allocated:\t%d\n",	state->total_num_bytes_allocated);
  printf ("GC Heap bytes collected:\t%d\n",	state->total_num_bytes_collected);
  printf ("Number of Finalizer Runs:\t%d\n",	state->num_finalizer_runs);
  printf ("Number of Allocated Classes:\t%d (%d since last GC)\n",
	  state->total_num_classes_allocated,
	  state->interval_num_classes_allocated);
  printf ("Number of Finalized Classes:\t%d\n", state->total_num_classes_finalized);
  printf ("Number of Allocated Arrays :\t%d (%d since last GC)\n",
	  state->total_num_arrays_allocated,
	  state->interval_num_arrays_allocated);
  printf ("Number of Finalized Arrays :\t%d\n", state->total_num_arrays_finalized);
  printf ("Number of Allocated Objects:\t%d (%d since last GC)\n",
	  state->total_num_objects_allocated,
	  state->interval_num_objects_allocated);
  printf ("Number of Finalized Objects:\t%d\n", state->total_num_objects_finalized);
  printf ("\n");

  MONITOR_exit(state->mon);
}

void
GC_addReference(JavaVM *vm, GC_obj *gc_obj)
{
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;
  int need_gc = 0;

  MONITOR_enter(state->mon);

  if (gc_obj->gc_flags & GC_FLAG_OBJECT)
    {
      state->total_num_objects_allocated++;
      state->interval_num_objects_allocated++;
    }
  else if (gc_obj->gc_flags & GC_FLAG_ARRAY)
    {
      state->total_num_arrays_allocated++;
      state->interval_num_arrays_allocated++;
    }
  else
    {
      state->total_num_classes_allocated++;
      state->interval_num_classes_allocated++;
    }

  state->total_num_bytes_allocated += gc_obj->size;
  state->interval_num_bytes_allocated += gc_obj->size;

  if (state->interval_num_bytes_allocated > MAX_BYTES)
    need_gc = 1;

#if notyet
  if (need_gc)
    GC_runCollector(vm);
#endif

  ENQUEUE(gc_obj, state->global_refs);

  MONITOR_exit(state->mon);
}

void
GC_addRoot(JavaVM *vm, GC_obj *gc_obj)
{
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;
  MONITOR_enter(state->mon);

  ENQUEUE(gc_obj, state->roots);

  MONITOR_exit(state->mon);
}

==> japhar-0.05.tar.gz: japhar-0.05/lib/libruntime/msgc.c <==
/* -*- Mode: C; c-file-style: "gnu" -*-
   msgc.c -- A simple, non-concurrent mark and sweep collector.
   Created: Chris Toshok <toshok@hungry.com>, 10-Aug-1997
*/
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "jniint.h"
#include "qop.h"
#include "gc.h"
#include "objects.h"
#include "arrays.h"
#include "array-class.h"
#include "ClazzFile.h"
#include "native-threads.h"

#include <stdlib.h>

#define MAX_BYTES 32000

GC_state*
GC_createGCState(JavaVM *vm)
{
  GC_state *state = NULL;

#if 0
  state = (GC_state*)calloc(1, sizeof(*state));

  state->mon = MONITOR_create();
  state->vm = vm;
#endif

  return state;
}

void
GC_deleteGCState(GC_state* state)
{
#if 0
  if (state->finalizer_q)
    GC_runFinalizers(state->vm);

  /* run the finalizers on all the roots. */

  /* destroy the monitor associated with this GC_state. */
  MONITOR_destroy(state->mon);

  /* free up the GC_state memory. */
  free(state);
#endif
}

void
GC_finalize(JavaVM *vm, GC_obj *gc_obj)
{
#if 0
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;

  if (gc_obj->gc_flags & GC_FLAG_CLASS
      && !state->class_gc_enabled)
    return;

  if (gc_obj->gc_flags & GC_FLAG_OBJECT)
    state->total_num_objects_finalized++;
  else if (gc_obj->gc_flags & GC_FLAG_ARRAY)
    state->total_num_arrays_finalized++;
  else
    state->total_num_classes_finalized++;

  state->total_num_bytes_collected += gc_obj->size;

  (*gc_obj->finalize)(gc_obj);
#endif
}

void
GC_deleteGlobal(JavaVM *vm, GC_obj *gc_obj)
{
#if 0
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;
  MONITOR_enter(state->mon);

  UNQUEUE(gc_obj, state->global_refs);
  GC_finalize(vm, gc_obj);

  MONITOR_exit(state->mon);
#endif
}

void
GC_deleteThreadLocal(JavaVM* vm, GC_obj *gc_obj)
{
#if 0
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;
  MONITOR_enter(state->mon);
  ENQUEUE(gc_obj, state->finalizer_q);
  MONITOR_exit(state->mon);
#endif
}

void
GC_runFinalizers(JavaVM *vm)
{
#if 0
  HungryJavaVM *hvm = (HungryJavaVM*)vm;
  GC_state *state = hvm->_gc_state;
  int classes, objs, arrays;

  classes =
    objs =
    arrays = 0;

  state->verbose_gc = hvm->_args.enableVerboseGC;
  state->class_gc_enabled = hvm->_args.enableClassGC;

  state->num_finalizer_runs++;

  if (state->verbose_gc)
    printf ("GC: finalizing.\n");

  while(state->finalizer_q)
    {
      GC_obj *next = state->finalizer_q->next;
      GC_obj *cur = state->finalizer_q;

      UNQUEUE(cur, state->finalizer_q);
      if (cur->gc_flags & GC_FLAG_OBJECT)
	objs ++;
      else if (cur->gc_flags & GC_FLAG_ARRAY)
	arrays++;
      else
	classes++;

      GC_finalize(vm, cur);

      state->finalizer_q = next;
    }

  if (state->verbose_gc)
    printf ("GC: finalizing complete (%d objs/%d classes/%d arrays)\n", objs, classes, arrays);
#endif
}

static void
gc_markStuff(JavaVM *vm, GC_obj *gc_obj)
{
#if 0
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;

  MONITOR_enter(state->mon);

  if (gc_obj->gc_flags & GC_FLAG_OBJECT)
    {
      japhar_object *obj = (japhar_object*)gc_obj;
      ClazzFile *clazz;
	  
      while (!IS_POST(obj))
	obj --;
      obj ++;
	  
      if (((GC_obj*)obj)->gc_flags & GC_FLAG_MARKED)
	return;
      ((GC_obj*)obj)->gc_flags |= GC_FLAG_MARKED;

      state->num_objects_reachable ++;

      while (!IS_POST(obj))
	{
	  int i;

	  /* mark the class of this thunk */
	  clazz = (ClazzFile*)STATIC_DATA(obj);
	  gc_markStuff(vm, (GC_obj*)clazz);

	  /* now go through the fields, and mark all the objects. */
	  for (i = 0; i < clazz->num_instance_fields; i ++)
	    {
	      if (INSTANCE_DATA(obj)->fields[i].tag == SIG_JOBJECT
		  && INSTANCE_DATA(obj)->fields[i].value.l != NULL)
		gc_markStuff(vm, (GC_obj*)INSTANCE_DATA(obj)->fields[i].value.l);
	    }

	  obj++;
	}
    }
  else if (gc_obj->gc_flags & GC_FLAG_ARRAY)
    {
      JavaStackItem type_item;

      if (gc_obj->gc_flags & GC_FLAG_MARKED)
	return;
      gc_obj->gc_flags |= GC_FLAG_MARKED;
	  
      state->num_arrays_reachable ++;

      get_instance_field((jobject)gc_obj,
			 STATIC_DATA((jobject)gc_obj)->fields[ARRAY_TYPE_INDEX],
			 &type_item);

      if (type_item.value.i == SIG_JOBJECT)
	{
	  JavaStackItem length_item;
	  JavaStackItem body_item;
	  int i;

	  get_instance_field((jobject)gc_obj,
			     STATIC_DATA((jobject)gc_obj)->fields[ARRAY_LENGTH_INDEX],
			     &length_item);
		  
	  get_instance_field((jobject)gc_obj,
			     STATIC_DATA((jobject)gc_obj)->fields[ARRAY_BODY_INDEX],
			     &body_item);

	  for (i = 0; i < length_item.value.i; i ++)
	    {
	      if (((jobject*)body_item.value.l)[i])
		gc_markStuff(vm, (GC_obj*)((jobject*)body_item.value.l)[i]);
	    }
	}
    }
  else if (gc_obj->gc_flags & GC_FLAG_CLASS)
    {
      ClazzFile *cf = (ClazzFile*)gc_obj;
      int i;
	 
      if (gc_obj->gc_flags & GC_FLAG_MARKED)
	{
	  MONITOR_exit(state->mon);

	  return;
	}

      gc_obj->gc_flags |= GC_FLAG_MARKED;
 
      state->num_classes_reachable ++;

      /* here we need to mark all the objects that we have in static
	 fields */
      for (i = 0; i < cf->num_fields; i ++)
	{
	  FieldStruct *field = cf->fields[i];
		  
	  if (field->access_flags & ACC_STATIC
	      && ((field->sig->any.tag == SIG_PRIM
		   && field->sig->prim.type == SIG_JOBJECT)
		  || field->sig->any.tag == SIG_ARRAY))
	    {
	      /* so, the field we're looking at is either an object or
		 an array. */
	      japhar_object *next_obj = cf->static_fields[ field->field_index ].value.l;
			  
	      gc_markStuff(vm, (GC_obj*)next_obj);
	    }
	}
    }

  MONITOR_exit(state->mon);
#endif
}

static void
gc_markRoots(JavaVM *vm)
{
#if 0
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;
  GC_obj *cur;

  for (cur = state->roots;
       cur != NULL;
       cur = cur->next)
    {
      gc_markStuff(vm, cur);
    }
#endif
}

static void
gc_markFromRegisters(JavaVM *vm)
{
}

static void
gc_markOperandStack(JavaVM *vm,
		    OpStack *opstack)
{
#if 0
  JavaStackItem *item;

  for (item = opstack->stack_top;
       item >= opstack->stack_bottom;
       item --)
    {
      if (item->tag == SIG_JOBJECT
	  && item->value.l != NULL)
	gc_markStuff(vm, (GC_obj*)item->value.l);
    }
#endif
}

static void
gc_markThreadReferences(JavaVM *vm)
{
#if 0
  HungryJavaVM* hvm = (HungryJavaVM*)vm;
  HungryJNIEnv* env;

  for (env = hvm->_envs;
       env != NULL;
       env = env->next)
    {
      GC_obj *ref;
	  
      for (ref = env->_local_refs;
	   ref != NULL;
	   ref = ref->next)
	gc_markStuff(vm, ref);

      gc_markOperandStack(vm, env->_java_info->op_stack);
    }
#endif
}

void
GC_runCollector(JavaVM *vm)
{
#if 0
  HungryJavaVM *hvm = (HungryJavaVM*)vm;
  GC_state *state = hvm->_gc_state;
  GC_obj *cur;
  int num_added_to_finalizer_q = 0;

  MONITOR_enter(state->mon);

  state->verbose_gc = hvm->_args.enableVerboseGC;
  state->class_gc_enabled = hvm->_args.enableClassGC;

  state->num_gcs++;

  state->num_objects_reachable =
    state->num_arrays_reachable =
    state->num_classes_reachable = 0;

  if (state->verbose_gc)
    printf ("<GC: starting (Mark and Sweep)>\n");

  /* first we go along our global ref's and mark the all unreached. */
  for (cur = state->global_refs;
       cur != NULL;
       cur = cur->next)
    {
      cur->gc_flags &= ~GC_FLAG_MARKED;
    }

  if (state->verbose_gc)
    printf ("  Mark phase.\n");
  /* mark phase */
  if (state->verbose_gc)
    printf ("    from roots.\n");
  gc_markRoots(vm);

  if (state->verbose_gc)
    printf ("    from registers.\n");
  gc_markFromRegisters(vm);

  if (state->verbose_gc)
    printf ("    from thread stacks.\n");
  gc_markThreadReferences(vm);

  if (state->verbose_gc)
    printf ("  Mark phase complete.\n");

  if (state->verbose_gc)
    {
      printf ("  %d objects reachable.\n",	state->num_objects_reachable);
      printf ("  %d arrays reachable.\n",	state->num_arrays_reachable);
      printf ("  %d classes reachable.\n",	state->num_classes_reachable);
    }

  if (state->verbose_gc)
    printf ("  Sweep phase.\n");
  /* sweep phase */
  cur = state->global_refs;
  while (cur != NULL)
    {
      if (!(cur->gc_flags & GC_FLAG_MARKED)
	  && (!(cur->gc_flags & GC_FLAG_CLASS) || state->class_gc_enabled))
	{
	  GC_obj *next = cur->next;
	  /* it's garbage. */
	  UNQUEUE(cur, state->global_refs);
	  ENQUEUE(cur, state->finalizer_q);

	  num_added_to_finalizer_q ++;
	  cur = next;
	}
      else
	cur = cur->next;
    }
  if (state->verbose_gc)
    printf ("  Sweep phase complete (%d added to finalizer queue.)\n",
	    num_added_to_finalizer_q);

  if (state->verbose_gc)
    GC_printStats(vm);

  GC_runFinalizers(vm);

  state->interval_num_bytes_allocated =
    state->interval_num_classes_allocated =
    state->interval_num_arrays_allocated =
    state->interval_num_objects_allocated = 0;

  MONITOR_exit(state->mon);
#endif
}

void
GC_printStats(JavaVM *vm)
{
#if 0
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;

  MONITOR_enter(state->mon);

  printf ("\n");
  printf ("Number of GC's:\t%d\n",	state->num_gcs);
  printf ("GC Heap bytes allocated:\t%d\n",	state->total_num_bytes_allocated);
  printf ("GC Heap bytes collected:\t%d\n",	state->total_num_bytes_collected);
  printf ("Number of Finalizer Runs:\t%d\n",	state->num_finalizer_runs);
  printf ("Number of Allocated Classes:\t%d (%d since last GC)\n",
	  state->total_num_classes_allocated,
	  state->interval_num_classes_allocated);
  printf ("Number of Finalized Classes:\t%d\n", state->total_num_classes_finalized);
  printf ("Number of Allocated Arrays :\t%d (%d since last GC)\n",
	  state->total_num_arrays_allocated,
	  state->interval_num_arrays_allocated);
  printf ("Number of Finalized Arrays :\t%d\n", state->total_num_arrays_finalized);
  printf ("Number of Allocated Objects:\t%d (%d since last GC)\n",
	  state->total_num_objects_allocated,
	  state->interval_num_objects_allocated);
  printf ("Number of Finalized Objects:\t%d\n", state->total_num_objects_finalized);
  printf ("\n");

  MONITOR_exit(state->mon);
#endif
}

void
GC_addReference(JavaVM *vm, GC_obj *gc_obj)
{
#if 0
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;
  int need_gc = 0;

  MONITOR_enter(state->mon);

  if (gc_obj->gc_flags & GC_FLAG_OBJECT)
    {
      state->total_num_objects_allocated++;
      state->interval_num_objects_allocated++;
    }
  else if (gc_obj->gc_flags & GC_FLAG_ARRAY)
    {
      state->total_num_arrays_allocated++;
      state->interval_num_arrays_allocated++;
    }
  else
    {
      state->total_num_classes_allocated++;
      state->interval_num_classes_allocated++;
    }

  state->total_num_bytes_allocated += gc_obj->size;
  state->interval_num_bytes_allocated += gc_obj->size;

  if (state->interval_num_bytes_allocated > MAX_BYTES)
    need_gc = 1;

#if notyet
  if (need_gc)
    GC_runCollector(vm);
#endif

  ENQUEUE(gc_obj, state->global_refs);

  MONITOR_exit(state->mon);
#endif
}

void
GC_addRoot(JavaVM *vm, GC_obj *gc_obj)
{
#if 0
  GC_state *state = ((HungryJavaVM*)vm)->_gc_state;
  MONITOR_enter(state->mon);

  ENQUEUE(gc_obj, state->roots);

  MONITOR_exit(state->mon);
#endif
}
